@@ -1,75 +1,34 @@
-## Copyright 2013-2014 Ray Holder
-##
-## Licensed under the Apache License, Version 2.0 (the "License");
-## you may not use this file except in compliance with the License.
-## You may obtain a copy of the License at
-##
-## http://www.apache.org/licenses/LICENSE-2.0
-##
-## Unless required by applicable law or agreed to in writing, software
-## distributed under the License is distributed on an "AS IS" BASIS,
-## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-## See the License for the specific language governing permissions and
-## limitations under the License.
-
-## --- The following is for portions of the "six" module ----------------------
-##
-## Copyright (c) 2010-2014 Benjamin Peterson
-##
-## Permission is hereby granted, free of charge, to any person obtaining a copy
-## of this software and associated documentation files (the "Software"), to deal
-## in the Software without restriction, including without limitation the rights
-## to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-## copies of the Software, and to permit persons to whom the Software is
-## furnished to do so, subject to the following conditions:
-##
-## The above copyright notice and this permission notice shall be included in all
-## copies or substantial portions of the Software.
-##
-## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-## IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-## FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-## AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-## LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-## OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-## SOFTWARE.
-## ----------------------------------------------------------------------------
-
+# Copyright 2013-2014 Ray Holder
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+# http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+import logging
 import random
 import sys
 import time
 import traceback
-
-# Python 3 compatibility hacks, pilfered from https://pypi.python.org/pypi/six/1.6.1
-PY3 = sys.version_info[0] == 3
-if PY3:
-    def reraise(tp, value, tb=None):
-        if value.__traceback__ is not tb:
-            raise value.with_traceback(tb)
-        raise value
-
-else:
-    def exec_(_code_, _globs_=None, _locs_=None):
-        """Execute code in a namespace."""
-        if _globs_ is None:
-            frame = sys._getframe(1)
-            _globs_ = frame.f_globals
-            if _locs_ is None:
-                _locs_ = frame.f_locals
-            del frame
-        elif _locs_ is None:
-            _locs_ = _globs_
-        exec("""exec _code_ in _globs_, _locs_""")
-
-
-    exec_("""def reraise(tp, value, tb=None):
-    raise tp, value, tb
-""")
+from functools import wraps
 
 # sys.maxint / 2, since Python 3.2 doesn't have a sys.maxint...
 MAX_WAIT = 1073741823
 
 
+def _retry_if_exception_of_type(retryable_types):
+    def _retry_if_exception_these_types(exception):
+        return isinstance(exception, retryable_types)
+
+    return _retry_if_exception_these_types
+
+
 def retry(*dargs, **dkw):
     """
     Decorator function that instantiates the Retrying object
@@ -78,7 +37,9 @@ def retry(*dargs, **dkw):
     """
     # support both @retry and @retry() as valid syntax
     if len(dargs) == 1 and callable(dargs[0]):
+
         def wrap_simple(f):
+            @wraps(f)
             def wrapped_f(*args, **kw):
                 return Retrying().call(f, *args, **kw)
 
@@ -87,7 +48,9 @@ def retry(*dargs, **dkw):
         return wrap_simple(dargs[0])
 
     else:
+
         def wrap(f):
+            @wraps(f)
             def wrapped_f(*args, **kw):
                 return Retrying(*dargs, **dkw).call(f, *args, **kw)
 
@@ -95,32 +58,80 @@ def retry(*dargs, **dkw):
 
         return wrap
 
+_default_logger = None
+_configured_null_logger = False
+
+def _pick_logger(logger=None):
+    # Factor this logic out into a smaller function so that `global` only needs to be here,
+    # not the large __init__ function.
+    global _default_logger, _configured_null_logger
+
+    if logger in (True, None):
+        if _default_logger is None:
+            _default_logger = logging.getLogger(__name__)
+        # Only add the null handler once, not every time we get the logger
+        if logger is None and not _configured_null_logger:
+            _configured_null_logger = True
+            _default_logger.addHandler(logging.NullHandler())
+            _default_logger.propagate = False
+        return _default_logger
+    else:  # Not None (and not True) -> must have supplied a logger. Just use that.
+        return logger
+
 
 class Retrying(object):
+    def __init__(
+        self,
+        stop=None,
+        wait=None,
+        stop_max_attempt_number=None,
+        stop_max_delay=None,
+        wait_fixed=None,
+        wait_random_min=None,
+        wait_random_max=None,
+        wait_incrementing_start=None,
+        wait_incrementing_increment=None,
+        wait_incrementing_max=None,
+        wait_exponential_multiplier=None,
+        wait_exponential_max=None,
+        retry_on_exception=None,
+        retry_on_result=None,
+        wrap_exception=False,
+        stop_func=None,
+        wait_func=None,
+        wait_jitter_max=None,
+        before_attempts=None,
+        after_attempts=None,
+        logger=None,
+    ):
 
-    def __init__(self,
-                 stop=None, wait=None,
-                 stop_max_attempt_number=None,
-                 stop_max_delay=None,
-                 wait_fixed=None,
-                 wait_random_min=None, wait_random_max=None,
-                 wait_incrementing_start=None, wait_incrementing_increment=None,
-                 wait_exponential_multiplier=None, wait_exponential_max=None,
-                 retry_on_exception=None,
-                 retry_on_result=None,
-                 wrap_exception=False,
-                 stop_func=None,
-                 wait_func=None):
-
-        self._stop_max_attempt_number = 5 if stop_max_attempt_number is None else stop_max_attempt_number
+        self._stop_max_attempt_number = (
+            5 if stop_max_attempt_number is None else stop_max_attempt_number
+        )
         self._stop_max_delay = 100 if stop_max_delay is None else stop_max_delay
         self._wait_fixed = 1000 if wait_fixed is None else wait_fixed
         self._wait_random_min = 0 if wait_random_min is None else wait_random_min
         self._wait_random_max = 1000 if wait_random_max is None else wait_random_max
-        self._wait_incrementing_start = 0 if wait_incrementing_start is None else wait_incrementing_start
-        self._wait_incrementing_increment = 100 if wait_incrementing_increment is None else wait_incrementing_increment
-        self._wait_exponential_multiplier = 1 if wait_exponential_multiplier is None else wait_exponential_multiplier
-        self._wait_exponential_max = MAX_WAIT if wait_exponential_max is None else wait_exponential_max
+        self._wait_incrementing_start = (
+            0 if wait_incrementing_start is None else wait_incrementing_start
+        )
+        self._wait_incrementing_increment = (
+            100 if wait_incrementing_increment is None else wait_incrementing_increment
+        )
+        self._wait_exponential_multiplier = (
+            1 if wait_exponential_multiplier is None else wait_exponential_multiplier
+        )
+        self._wait_exponential_max = (
+            MAX_WAIT if wait_exponential_max is None else wait_exponential_max
+        )
+        self._wait_incrementing_max = (
+            MAX_WAIT if wait_incrementing_max is None else wait_incrementing_max
+        )
+        self._wait_jitter_max = 0 if wait_jitter_max is None else wait_jitter_max
+        self._before_attempts = before_attempts
+        self._after_attempts = after_attempts
+
+        self._logger = _pick_logger(logger)
 
         # TODO add chaining of stop behaviors
         # stop behavior
@@ -135,7 +146,9 @@ class Retrying(object):
             self.stop = stop_func
 
         elif stop is None:
-            self.stop = lambda attempts, delay: any(f(attempts, delay) for f in stop_funcs)
+            self.stop = lambda attempts, delay: any(
+                f(attempts, delay) for f in stop_funcs
+            )
 
         else:
             self.stop = getattr(self, stop)
@@ -149,7 +162,10 @@ class Retrying(object):
         if wait_random_min is not None or wait_random_max is not None:
             wait_funcs.append(self.random_sleep)
 
-        if wait_incrementing_start is not None or wait_incrementing_increment is not None:
+        if (
+            wait_incrementing_start is not None
+            or wait_incrementing_increment is not None
+        ):
             wait_funcs.append(self.incrementing_sleep)
 
         if wait_exponential_multiplier is not None or wait_exponential_max is not None:
@@ -159,7 +175,9 @@ class Retrying(object):
             self.wait = wait_func
 
         elif wait is None:
-            self.wait = lambda attempts, delay: max(f(attempts, delay) for f in wait_funcs)
+            self.wait = lambda attempts, delay: max(
+                f(attempts, delay) for f in wait_funcs
+            )
 
         else:
             self.wait = getattr(self, wait)
@@ -168,9 +186,13 @@ class Retrying(object):
         if retry_on_exception is None:
             self._retry_on_exception = self.always_reject
         else:
+            # this allows for providing a tuple of exception types that
+            # should be allowed to retry on, and avoids having to create
+            # a callback that does the same thing
+            if isinstance(retry_on_exception, (tuple, Exception)):
+                retry_on_exception = _retry_if_exception_of_type(retry_on_exception)
             self._retry_on_exception = retry_on_exception
 
-        # TODO simplify retrying by Exception types
         # retry on result filter
         if retry_on_result is None:
             self._retry_on_result = self.never_reject
@@ -187,7 +209,8 @@ class Retrying(object):
         """Stop after the time from the first attempt >= stop_max_delay."""
         return delay_since_first_attempt_ms >= self._stop_max_delay
 
-    def no_sleep(self, previous_attempt_number, delay_since_first_attempt_ms):
+    @staticmethod
+    def no_sleep(previous_attempt_number, delay_since_first_attempt_ms):
         """Don't sleep at all before retrying."""
         return 0
 
@@ -204,13 +227,17 @@ class Retrying(object):
         Sleep an incremental amount of time after each attempt, starting at
         wait_incrementing_start and incrementing by wait_incrementing_increment
         """
-        result = self._wait_incrementing_start + (self._wait_incrementing_increment * (previous_attempt_number - 1))
+        result = self._wait_incrementing_start + (
+            self._wait_incrementing_increment * (previous_attempt_number - 1)
+        )
+        if result > self._wait_incrementing_max:
+            result = self._wait_incrementing_max
         if result < 0:
             result = 0
         return result
 
     def exponential_sleep(self, previous_attempt_number, delay_since_first_attempt_ms):
-        exp = 2 ** previous_attempt_number
+        exp = 2**previous_attempt_number
         result = self._wait_exponential_multiplier * exp
         if result > self._wait_exponential_max:
             result = self._wait_exponential_max
@@ -218,10 +245,12 @@ class Retrying(object):
             result = 0
         return result
 
-    def never_reject(self, result):
+    @staticmethod
+    def never_reject(result):
         return False
 
-    def always_reject(self, result):
+    @staticmethod
+    def always_reject(result):
         return True
 
     def should_reject(self, attempt):
@@ -237,15 +266,22 @@ class Retrying(object):
         start_time = int(round(time.time() * 1000))
         attempt_number = 1
         while True:
+            if self._before_attempts:
+                self._before_attempts(attempt_number)
+
             try:
                 attempt = Attempt(fn(*args, **kwargs), attempt_number, False)
-            except:
+            except Exception:
                 tb = sys.exc_info()
                 attempt = Attempt(tb, attempt_number, True)
 
             if not self.should_reject(attempt):
                 return attempt.get(self._wrap_exception)
 
+            self._logger.warning(attempt)
+            if self._after_attempts:
+                self._after_attempts(attempt_number)
+
             delay_since_first_attempt_ms = int(round(time.time() * 1000)) - start_time
             if self.stop(attempt_number, delay_since_first_attempt_ms):
                 if not self._wrap_exception and attempt.has_exception:
@@ -255,6 +291,10 @@ class Retrying(object):
                     raise RetryError(attempt)
             else:
                 sleep = self.wait(attempt_number, delay_since_first_attempt_ms)
+                if self._wait_jitter_max:
+                    jitter = random.random() * self._wait_jitter_max
+                    sleep = sleep + max(0, jitter)
+                self._logger.info(f"Retrying in {sleep / 1000.0:.2f} seconds.")
                 time.sleep(sleep / 1000.0)
 
             attempt_number += 1
@@ -282,15 +322,16 @@ class Attempt(object):
             if wrap_exception:
                 raise RetryError(self)
             else:
-                reraise(self.value[0], self.value[1], self.value[2])
+                exc_type, exc, tb = self.value
+                raise exc.with_traceback(tb)
         else:
             return self.value
 
     def __repr__(self):
         if self.has_exception:
-            return "Attempts: {0}, Error:\n{1}".format(self.attempt_number, "".join(traceback.format_tb(self.value[2])))
+            return f"Attempts: {self.attempt_number}, Error:\n{''.join(traceback.format_tb(self.value[2]))}"
         else:
-            return "Attempts: {0}, Value: {1}".format(self.attempt_number, self.value)
+            return f"Attempts: {self.attempt_number}, Value: {self.value}"
 
 
 class RetryError(Exception):
@@ -302,4 +343,4 @@ class RetryError(Exception):
         self.last_attempt = last_attempt
 
     def __str__(self):
-        return "RetryError[{0}]".format(self.last_attempt)
+        return f"RetryError[{self.last_attempt}]"
